==> create-fly-1.21.8-6.0.8-2-sources.jar: com/zurrtum/create/content/logistics/packagerLink/LogisticallyLinkedBehaviour.java <==
package com.zurrtum.create.content.logistics.packagerLink;

import com.google.common.cache.Cache;
import com.zurrtum.create.Create;
import com.zurrtum.create.catnip.data.Pair;
import com.zurrtum.create.content.logistics.packager.IdentifiedInventory;
import com.zurrtum.create.content.logistics.packager.InventorySummary;
import com.zurrtum.create.content.logistics.packager.PackagerBlockEntity;
import com.zurrtum.create.content.logistics.packager.PackagingRequest;
import com.zurrtum.create.foundation.blockEntity.SmartBlockEntity;
import com.zurrtum.create.foundation.blockEntity.behaviour.BehaviourType;
import com.zurrtum.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.zurrtum.create.foundation.utility.TickBasedCache;
import com.zurrtum.create.infrastructure.component.PackageOrderWithCrafts;
import com.zurrtum.create.infrastructure.items.ItemStackHandler;
import org.apache.commons.lang3.mutable.MutableBoolean;
import org.jetbrains.annotations.Nullable;

import java.lang.ref.WeakReference;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;
import net.minecraft.class_11368;
import net.minecraft.class_11372;
import net.minecraft.class_124;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_2561;
import net.minecraft.class_4208;
import net.minecraft.class_4844;

public class LogisticallyLinkedBehaviour extends BlockEntityBehaviour<SmartBlockEntity> {

    public static final BehaviourType<LogisticallyLinkedBehaviour> TYPE = new BehaviourType<>();

    public static final AtomicInteger LINK_ID_GENERATOR = new AtomicInteger();
    public int linkId; // Runtime context, not saved to disk

    public int redstonePower;
    public UUID freqId;

    private boolean addedGlobally = false;
    private boolean loadedGlobally = false;
    private boolean global;

    public enum RequestType {
        RESTOCK,
        REDSTONE,
        PLAYER
    }

    private static final Cache<UUID, Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>>> LINKS = new TickBasedCache<>(20, true);

    private static final Cache<UUID, Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>>> CLIENT_LINKS = new TickBasedCache<>(20, true, true);

    public LogisticallyLinkedBehaviour(SmartBlockEntity be, boolean global) {
        super(be);
        this.global = global;
        linkId = LINK_ID_GENERATOR.getAndIncrement();
        freqId = UUID.randomUUID();
    }

    public static Collection<LogisticallyLinkedBehaviour> getAllPresent(UUID freq, boolean sortByPriority) {
        return getAllPresent(freq, sortByPriority, false);
    }

    public static Collection<LogisticallyLinkedBehaviour> getAllPresent(UUID freq, boolean sortByPriority, boolean clientSide) {
        Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = (clientSide ? CLIENT_LINKS : LINKS).getIfPresent(freq);
        if (cache == null)
            return Collections.emptyList();
        Stream<LogisticallyLinkedBehaviour> stream = new LinkedList<>(cache.asMap().values()).stream().map(WeakReference::get)
            .filter(LogisticallyLinkedBehaviour::isValidLink);

        if (sortByPriority)
            stream = stream.sorted((e1, e2) -> Integer.compare(e1.redstonePower, e2.redstonePower));

        return stream.toList();
    }

    public static void keepAlive(LogisticallyLinkedBehaviour behaviour) {
        boolean onClient = behaviour.blockEntity.method_10997().field_9236;
        if (behaviour.redstonePower == 15)
            return;
        try {
            Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = (onClient ? CLIENT_LINKS : LINKS).get(
                behaviour.freqId,
                () -> new TickBasedCache<>(400, false)
            );

            if (cache == null)
                return;

            WeakReference<LogisticallyLinkedBehaviour> reference = cache.get(behaviour.linkId, () -> new WeakReference<>(behaviour));
            cache.put(behaviour.linkId, reference.get() != behaviour ? new WeakReference<>(behaviour) : reference);

        } catch (ExecutionException e) {
            e.printStackTrace();
        }
    }

    public static void remove(LogisticallyLinkedBehaviour behaviour) {
        Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = LINKS.getIfPresent(behaviour.freqId);
        if (cache != null)
            cache.invalidate(behaviour.linkId);
    }

    //

    @Override
    public void unload() {
        if (loadedGlobally && global && getWorld() != null)
            Create.LOGISTICS.linkInvalidated(freqId, getGlobalPos());
        super.unload();
        remove(this);
    }

    @Override
    public void lazyTick() {
        keepAlive(this);
    }

    @Override
    public void initialize() {
        super.initialize();
        if (getWorld().field_9236)
            return;

        if (!loadedGlobally && global) {
            loadedGlobally = true;
            Create.LOGISTICS.linkLoaded(freqId, getGlobalPos());
        }

        if (!addedGlobally && global) {
            addedGlobally = true;
            blockEntity.method_5431();
            if (blockEntity instanceof PackagerLinkBlockEntity plbe)
                Create.LOGISTICS.linkAdded(freqId, getGlobalPos(), plbe.placedBy);
        }

    }

    private class_4208 getGlobalPos() {
        return class_4208.method_19443(getWorld().method_27983(), getPos());
    }

    @Override
    public void destroy() {
        super.destroy();
        if (addedGlobally && global && getWorld() != null)
            Create.LOGISTICS.linkRemoved(freqId, getGlobalPos());
    }

    public void redstonePowerChanged(int power) {
        if (power == redstonePower)
            return;
        redstonePower = power;
        blockEntity.method_5431();

        if (power == 15)
            remove(this);
        else
            keepAlive(this);
    }

    public Pair<PackagerBlockEntity, PackagingRequest> processRequest(
        class_1799 stack,
        int amount,
        String address,
        int linkIndex,
        MutableBoolean finalLink,
        int orderId,
        @Nullable PackageOrderWithCrafts context,
        @Nullable IdentifiedInventory ignoredHandler
    ) {

        if (blockEntity instanceof PackagerLinkBlockEntity plbe)
            return plbe.processRequest(stack, amount, address, linkIndex, finalLink, orderId, context, ignoredHandler);

        return null;
    }

    public InventorySummary getSummary(@Nullable IdentifiedInventory ignoredHandler) {
        if (blockEntity instanceof PackagerLinkBlockEntity plbe)
            return plbe.fetchSummaryFromPackager(ignoredHandler);
        return InventorySummary.EMPTY;
    }

    public void deductFromAccurateSummary(ItemStackHandler packageContents) {
        InventorySummary summary = LogisticsManager.ACCURATE_SUMMARIES.getIfPresent(freqId);
        if (summary == null)
            return;
        for (int i = 0, size = packageContents.method_5439(); i < size; i++) {
            class_1799 orderedStack = packageContents.method_5438(i);
            if (orderedStack.method_7960())
                continue;
            summary.add(orderedStack, -Math.min(summary.getCountOf(orderedStack), orderedStack.method_7947()));
        }
    }

    //

    public boolean mayInteract(class_1657 player) {
        return Create.LOGISTICS.mayInteract(freqId, player);
    }

    public boolean mayInteractMessage(class_1657 player) {
        boolean mayInteract = Create.LOGISTICS.mayInteract(freqId, player);
        if (!mayInteract)
            player.method_7353(class_2561.method_43471("create.logistically_linked.protected").method_27692(class_124.field_1061), true);
        return mayInteract;
    }

    public boolean mayAdministrate(class_1657 player) {
        return Create.LOGISTICS.mayAdministrate(freqId, player);
    }

    public static boolean isValidLink(LogisticallyLinkedBehaviour link) {
        return link != null && !link.blockEntity.method_11015() && !link.blockEntity.isChunkUnloaded();
    }

    @Override
    public boolean isSafeNBT() {
        return true;
    }

    @Override
    public void writeSafe(class_11372 view) {
        view.method_71468("Freq", class_4844.field_25122, freqId);
    }

    @Override
    public void write(class_11372 view, boolean clientPacket) {
        super.write(view, clientPacket);
        view.method_71468("Freq", class_4844.field_25122, freqId);
        view.method_71465("Power", redstonePower);
        view.method_71472("Added", addedGlobally);
    }

    @Override
    public void read(class_11368 view, boolean clientPacket) {
        super.read(view, clientPacket);
        freqId = view.method_71426("Freq", class_4844.field_25122).orElse(null);
        redstonePower = view.method_71424("Power", 0);
        addedGlobally = view.method_71433("Added", false);
    }

    @Override
    public BehaviourType<?> getType() {
        return TYPE;
    }

}

==> create-fly-1.21.10-6.0.8-5-sources.jar: com/zurrtum/create/content/logistics/packagerLink/LogisticallyLinkedBehaviour.java <==
package com.zurrtum.create.content.logistics.packagerLink;

import com.google.common.cache.Cache;
import com.zurrtum.create.Create;
import com.zurrtum.create.catnip.data.Pair;
import com.zurrtum.create.content.logistics.packager.IdentifiedInventory;
import com.zurrtum.create.content.logistics.packager.InventorySummary;
import com.zurrtum.create.content.logistics.packager.PackagerBlockEntity;
import com.zurrtum.create.content.logistics.packager.PackagingRequest;
import com.zurrtum.create.foundation.blockEntity.SmartBlockEntity;
import com.zurrtum.create.foundation.blockEntity.behaviour.BehaviourType;
import com.zurrtum.create.foundation.blockEntity.behaviour.BlockEntityBehaviour;
import com.zurrtum.create.foundation.utility.TickBasedCache;
import com.zurrtum.create.infrastructure.component.PackageOrderWithCrafts;
import com.zurrtum.create.infrastructure.items.ItemStackHandler;
import org.apache.commons.lang3.mutable.MutableBoolean;
import org.jetbrains.annotations.Nullable;

import java.lang.ref.WeakReference;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedList;
import java.util.UUID;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Stream;
import net.minecraft.class_11368;
import net.minecraft.class_11372;
import net.minecraft.class_124;
import net.minecraft.class_1657;
import net.minecraft.class_1799;
import net.minecraft.class_2561;
import net.minecraft.class_4208;
import net.minecraft.class_4844;

public class LogisticallyLinkedBehaviour extends BlockEntityBehaviour<SmartBlockEntity> {

    public static final BehaviourType<LogisticallyLinkedBehaviour> TYPE = new BehaviourType<>();

    public static final AtomicInteger LINK_ID_GENERATOR = new AtomicInteger();
    public int linkId; // Runtime context, not saved to disk

    public int redstonePower;
    public UUID freqId;

    private boolean addedGlobally = false;
    private boolean loadedGlobally = false;
    private boolean global;

    public enum RequestType {
        RESTOCK,
        REDSTONE,
        PLAYER
    }

    private static final Cache<UUID, Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>>> LINKS = new TickBasedCache<>(20, true);

    private static final Cache<UUID, Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>>> CLIENT_LINKS = new TickBasedCache<>(20, true, true);

    public LogisticallyLinkedBehaviour(SmartBlockEntity be, boolean global) {
        super(be);
        this.global = global;
        linkId = LINK_ID_GENERATOR.getAndIncrement();
        freqId = UUID.randomUUID();
    }

    public static Collection<LogisticallyLinkedBehaviour> getAllPresent(UUID freq, boolean sortByPriority) {
        return getAllPresent(freq, sortByPriority, false);
    }

    public static Collection<LogisticallyLinkedBehaviour> getAllPresent(UUID freq, boolean sortByPriority, boolean clientSide) {
        Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = (clientSide ? CLIENT_LINKS : LINKS).getIfPresent(freq);
        if (cache == null)
            return Collections.emptyList();
        Stream<LogisticallyLinkedBehaviour> stream = new LinkedList<>(cache.asMap().values()).stream().map(WeakReference::get)
            .filter(LogisticallyLinkedBehaviour::isValidLink);

        if (sortByPriority)
            stream = stream.sorted((e1, e2) -> Integer.compare(e1.redstonePower, e2.redstonePower));

        return stream.toList();
    }

    public static void keepAlive(LogisticallyLinkedBehaviour behaviour) {
        boolean onClient = behaviour.blockEntity.method_10997().method_8608();
        if (behaviour.redstonePower == 15)
            return;
        try {
            Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = (onClient ? CLIENT_LINKS : LINKS).get(
                behaviour.freqId,
                () -> new TickBasedCache<>(400, false)
            );

            if (cache == null)
                return;

            WeakReference<LogisticallyLinkedBehaviour> reference = cache.get(behaviour.linkId, () -> new WeakReference<>(behaviour));
            cache.put(behaviour.linkId, reference.get() != behaviour ? new WeakReference<>(behaviour) : reference);

        } catch (ExecutionException e) {
            e.printStackTrace();
        }
    }

    public static void remove(LogisticallyLinkedBehaviour behaviour) {
        Cache<Integer, WeakReference<LogisticallyLinkedBehaviour>> cache = LINKS.getIfPresent(behaviour.freqId);
        if (cache != null)
            cache.invalidate(behaviour.linkId);
    }

    //

    @Override
    public void unload() {
        if (loadedGlobally && global && getWorld() != null)
            Create.LOGISTICS.linkInvalidated(freqId, getGlobalPos());
        super.unload();
        remove(this);
    }

    @Override
    public void lazyTick() {
        keepAlive(this);
    }

    @Override
    public void initialize() {
        super.initialize();
        if (getWorld().method_8608())
            return;

        if (!loadedGlobally && global) {
            loadedGlobally = true;
            Create.LOGISTICS.linkLoaded(freqId, getGlobalPos());
        }

        if (!addedGlobally && global) {
            addedGlobally = true;
            blockEntity.method_5431();
            if (blockEntity instanceof PackagerLinkBlockEntity plbe)
                Create.LOGISTICS.linkAdded(freqId, getGlobalPos(), plbe.placedBy);
        }

    }

    private class_4208 getGlobalPos() {
        return class_4208.method_19443(getWorld().method_27983(), getPos());
    }

    @Override
    public void destroy() {
        super.destroy();
        if (addedGlobally && global && getWorld() != null)
            Create.LOGISTICS.linkRemoved(freqId, getGlobalPos());
    }

    public void redstonePowerChanged(int power) {
        if (power == redstonePower)
            return;
        redstonePower = power;
        blockEntity.method_5431();

        if (power == 15)
            remove(this);
        else
            keepAlive(this);
    }

    public Pair<PackagerBlockEntity, PackagingRequest> processRequest(
        class_1799 stack,
        int amount,
        String address,
        int linkIndex,
        MutableBoolean finalLink,
        int orderId,
        @Nullable PackageOrderWithCrafts context,
        @Nullable IdentifiedInventory ignoredHandler
    ) {

        if (blockEntity instanceof PackagerLinkBlockEntity plbe)
            return plbe.processRequest(stack, amount, address, linkIndex, finalLink, orderId, context, ignoredHandler);

        return null;
    }

    public InventorySummary getSummary(@Nullable IdentifiedInventory ignoredHandler) {
        if (blockEntity instanceof PackagerLinkBlockEntity plbe)
            return plbe.fetchSummaryFromPackager(ignoredHandler);
        return InventorySummary.EMPTY;
    }

    public void deductFromAccurateSummary(ItemStackHandler packageContents) {
        InventorySummary summary = LogisticsManager.ACCURATE_SUMMARIES.getIfPresent(freqId);
        if (summary == null)
            return;
        for (int i = 0, size = packageContents.method_5439(); i < size; i++) {
            class_1799 orderedStack = packageContents.method_5438(i);
            if (orderedStack.method_7960())
                continue;
            summary.add(orderedStack, -Math.min(summary.getCountOf(orderedStack), orderedStack.method_7947()));
        }
    }

    //

    public boolean mayInteract(class_1657 player) {
        return Create.LOGISTICS.mayInteract(freqId, player);
    }

    public boolean mayInteractMessage(class_1657 player) {
        boolean mayInteract = Create.LOGISTICS.mayInteract(freqId, player);
        if (!mayInteract)
            player.method_7353(class_2561.method_43471("create.logistically_linked.protected").method_27692(class_124.field_1061), true);
        return mayInteract;
    }

    public boolean mayAdministrate(class_1657 player) {
        return Create.LOGISTICS.mayAdministrate(freqId, player);
    }

    public static boolean isValidLink(LogisticallyLinkedBehaviour link) {
        return link != null && !link.blockEntity.method_11015() && !link.blockEntity.isChunkUnloaded();
    }

    @Override
    public boolean isSafeNBT() {
        return true;
    }

    @Override
    public void writeSafe(class_11372 view) {
        view.method_71468("Freq", class_4844.field_25122, freqId);
    }

    @Override
    public void write(class_11372 view, boolean clientPacket) {
        super.write(view, clientPacket);
        view.method_71468("Freq", class_4844.field_25122, freqId);
        view.method_71465("Power", redstonePower);
        view.method_71472("Added", addedGlobally);
    }

    @Override
    public void read(class_11368 view, boolean clientPacket) {
        super.read(view, clientPacket);
        freqId = view.method_71426("Freq", class_4844.field_25122).orElse(null);
        redstonePower = view.method_71424("Power", 0);
        addedGlobally = view.method_71433("Added", false);
    }

    @Override
    public BehaviourType<?> getType() {
        return TYPE;
    }

}
